
/* clib.q: Q's system module */

/* This file is part of the Q programming system.

   The Q programming system is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option) any
   later version.

   The Q programming system is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

import stdlib, string;

/* Additional string functions. Some trivial stuff from the C library you
   might have been missing in Q. */

/* character predicates */

public extern islower C, isupper C, isalpha C, isdigit C, isxdigit C,
  isalnum C, ispunct C, isspace C, isgraph C, isprint C, iscntrl C, isascii C;

/* convert a string to lower- or uppercase */

public extern tolower S, toupper S;

/* Enhanced file functions. The fopen version provided here handles the `+'
   flag in mode strings, thus enabling you to open files for both reading and
   writing. The freopen function reopens an existing file object on another
   file. The ftell/fseek functions are used for file positioning. The tmpnam
   function returns a unique name for a temporary file, and tmpfile constructs
   a temporary file opened in "w+b" mode which will be deleted automatically
   when it is closed. */

public extern ::fopen NAME MODE, freopen NAME MODE F;
public extern ftell F, fseek F N;
public extern tmpnam, tmpfile;

/* Binary file I/O. These operations are used to read/write binary numeric
   data (integers and floating point values) from/to files opened in binary
   mode. A function to read a string of a given length from a file is also
   provided. */

public extern intsize X, uintsize X; // byte size of an integer

/* F is the file to be read/written. */

public extern readint F COUNT;	// read a signed integer (COUNT bytes)
public extern readuint F COUNT;	// read an unsigned integer (COUNT bytes)
public extern readfloat F;	// read a floating point value (8 bytes)
public extern readcfloat F;	// read a small floating point value (4 bytes)

public extern readstr F COUNT;	// read a string (COUNT bytes); NOTE: the
				// string will be truncated at the first 0 byte

public extern writeint F COUNT X;
				// write an integer (COUNT bytes)
public extern writefloat F X;	// write a floating point value (8 bytes)
public extern writecfloat F X;	// write a small floating point value (4 bytes)

/* C-style formatted I/O. These functions provide an interface to the C printf
   and scanf routines. Arguments to the printf routines and the results of the
   scanf routines are encoded as tuples or single non-tuple value (if only one
   item is read/written). */

public extern printf FORMAT ARGS, fprintf F FORMAT ARGS, sprintf FORMAT ARGS;
public extern scanf FORMAT, fscanf F FORMAT, sscanf S FORMAT;

/* File and system utility functions. These functions provide the same
   functionality as their C counterparts. Note that not all functions will
   work on all systems. */

public extern rename OLD NEW, unlink NAME;
					// rename and delete files
public extern getcwd, chdir NAME;	// access/change the working directory
public extern mkdir NAME, rmdir NAME;	// create and remove directories
public extern mkfifo NAME;		// create a named pipe (UNIX only)

public extern umask N;			// set/get file creation mask

public extern stat NAME, lstat NAME;	// file and link information

public extern system CMD;		// execute command using the shell

public extern fork;			// fork a child process
public extern exec PROG ARGS;		// execute program
public extern spawn PROG ARGS;		// execute program in child process
public extern exit N;			// exit process with given exit code
public extern kill SIG PID;		// send signal to process
public extern getpid;			// current process id
public extern wait;			// wait for any child process
public extern waitpid PID;		// wait for given child process
public extern status PID;		// current status of process

public extern getenv NAME, setenv NAME VAL;
					// get and set environment variables

/* Access routines for the components of the tuples returned by the stat/lstat
   functions. */

public st_dev STAT, st_ino STAT, st_mode STAT, st_nlink STAT, st_uid STAT,
  st_gid STAT, st_rdev STAT, st_size STAT, st_atime STAT, st_mtime STAT,
  st_ctime STAT;

st_dev STAT				= STAT!0;
st_ino STAT				= STAT!1;
st_mode STAT				= STAT!2;
st_nlink STAT				= STAT!3;
st_uid STAT				= STAT!4;
st_gid STAT				= STAT!5;
st_rdev STAT				= STAT!6;
st_size STAT				= STAT!7;
st_atime STAT				= STAT!8;
st_mtime STAT				= STAT!9;
st_ctime STAT				= STAT!10;

/* Routines to interpret the exit codes returned by the spawn/system/wait/
   waitpid/status functions. */

public extern isactive STATUS;
// process is active
public extern isexited STATUS, exitstatus STATUS;
// process has exited normally, get its exit code
public extern issignaled STATUS, termsig STATUS;
// process was terminated by signal, get the signal number
public extern isstopped STATUS, stopsig STATUS;
// process was stopped by signal, get the signal number

/* Multithreading. These operations are in close correspondence with POSIX
   1003.1b. However, some operations are named differently, and since there
   are no mutable data structures in Q, semaphores provide the extra
   functionality of sending data from one thread to another. Moreover,
   semaphores can also be bounded. Mutexes are also supported, mostly for the
   purpose of handling critical sections involving operations with
   side-effects (I/O etc.). Mutexes are *not* required to make conditions work
   since these have their own internal mutex handling. For more information on
   POSIX threads, please refer to the corresponding section in the UNIX
   manual. */

/* These functions will only work as advertised if the interpreter has been
   built with POSIX thread support (--with-pthread). It must also be noted
   that in the current implementation the interpreter effectively serializes
   multithreaded scripts on the reduction level and thus user-level threads
   cannot really take advantage of multi-processor machines. */

/* Thread creation and management. */

public extern type Thread;		// thread handle type

public isthread THREAD;			// check for thread objects

isthread _:Thread			= true;
isthread _				= false otherwise;

public extern thread_no THREAD;		// thread number
public extern this_thread;		// handle of the current thread

(T1:Thread = T2:Thread)			= thread_no T1 = thread_no T2;
(T1:Thread < T2:Thread)			= thread_no T1 < thread_no T2;
(T1:Thread > T2:Thread)			= thread_no T1 > thread_no T2;
(T1:Thread <> T2:Thread)		= thread_no T1 <> thread_no T2;
(T1:Thread <= T2:Thread)		= thread_no T1 <= thread_no T2;
(T1:Thread >= T2:Thread)		= thread_no T1 >= thread_no T2;

/* The main thread (private). Don't remove this definition, it is needed to
   prevent the main thread handle from being garbage collected! */

private var MAIN_THREAD;
def MAIN_THREAD				= this_thread;

public extern special thread X;		// start a thread, return its handle
public extern return X;			// terminate thread, return X as result
public extern cancel THREAD;		// cancel THREAD
public extern result THREAD;		// wait for THREAD, return its result
public extern yield;			// allow context switch

public extern active THREAD;		// check if THREAD is active
public extern canceled THREAD;		// check if THREAD was canceled

public extern schedule THREAD POL PRIO;	// change scheduling parameters

/* Mutexes. */

public extern type Mutex;		// mutex type

public ismutex MUTEX;			// check for mutex objects

ismutex _:Mutex				= true;
ismutex _				= false otherwise;

public extern mutex;			// standard, i.e., fast mutex object
public extern errorchecking_mutex;	// error checking mutex object
public extern recursive_mutex;		// recursive mutex object

public extern lock MUTEX;		// lock MUTEX
public extern unlock MUTEX;		// unlock MUTEX
public extern try MUTEX;		// try MUTEX

/* Conditions. */

public extern type Cond;		// condition type

public iscond COND;			// check for condition objects

iscond _:Cond				= true;
iscond _				= false otherwise;

public extern cond;			// new condition object

public extern signal COND;		// signal COND
public extern broadcast COND;		// broadcast COND
public extern suspend COND;		// wait for COND, or (COND,TIME)

/* Semaphores (semaphore queues, actually). */

public extern type Sem;			// semaphore type

public issem SEM;			// check for semaphore objects

issem _:Sem				= true;
issem _					= false otherwise;

public extern sem;			// semaphore object
public extern bounded_sem MAX;		// bounded semaphore object

public extern post SEM X;		// enqueue a value
public extern get SEM;			// dequeue a value
public extern try SEM;			// dequeue a value, fail if none

public extern get_size SEM;		// get the current queue size
public extern get_bound SEM;		// get the max queue size (0 if none)

#SEM:Sem				= get_size SEM;

/* Time functions. Return the active timezone, and a readable representation
   of any integer or floating point time value denoting the number of seconds
   elapsed since the "epoch", 00:00:00 on January 1, 1970, Coordinated
   Universal Time (UTC). Also available: current cpu time in "clock ticks",
   and the number of ticks per sec. */

public extern timezone, ctime T, gmtime T;
public extern clock, clocks_per_sec;

/* Filename globbing using the shell's wildcard syntax (*, ? etc.). */

public extern fnmatch PATTERN S;	// check whether S matches PATTERN
public extern glob PATTERN;		// return the list of all filenames
					// matching PATTERN

/* Regular expression matching using "extended" (egrep-like) syntax as defined
   by POSIX 1003.2/D11.2. */

/* 1. Low-level interface. The following functions are directly implemented in
   C using the GNU regex library. The regmatch function searches for the first
   match, regnext for the next, and regdone terminates a global search still
   in progress. The OPTS string allows you to specify various options for the
   search. In particular, "g" denotes a global, "i" a case-insensitive, and
   "n" a "multi-line" search; see the documentation for further details. */

public extern regmatch OPTS REGEX S, regnext, regdone;

/* 2. High-level interface. The regex function evaluates, for each match of
   the given regular expression in the given string, the special EXPR
   argument, and returns the collection of all results as a list. The OPTS
   argument has the same meaning as with the low-level functions. In
   particular, if the "g" option is omitted, then only the first match will be
   reported, if any. */

public special regex ~OPTS ~REGEX ~S EXPR, regex_next EXPR;

private special check ~P X Y;

regex OPTS:String REGEX:String S:String EXPR
				= check (regmatch OPTS REGEX S)
				  [EXPR|regex_next EXPR] [];

regex_next EXPR			= check regnext
				  [EXPR|regex_next EXPR] [];

check P:Bool X Y		= X if P;
				= Y otherwise;
check P X Y			= P otherwise;

/* 3. Match state information. These functions are typically invoked after
   regmatch, regnext, or in the EXPR argument of regex, to return information
   about the current match. The match state is maintained on a hidden stack
   manipulated with the regmatch/regnext/regdone functions, hence multiple
   nested searches are possible. */

public extern regstart, regskip, reg N, regpos N, regend N, regs;

/* An expression of the form `regerr MSG', where MSG is the error message, is
   used to return abnormal error conditions such as bad regular expression
   syntax. You can redefine `regerr' as appropriate for your application. */

public regerr MSG;

/* Additional integer functions from the GMP library. */

/* exact powers, integer parts of roots, powers/inverses mod K */

public extern pow M N, root M N, intsqrt M;
public extern powmod K M N, invmod K M;

/* number-theoretic functions: probabilistic prime test, gcd, lcm,
   factorization helpers, and Jacobi symbol */

public extern isprime N;
public extern gcd M N, lcm M N, fact M N, rem M N;
public extern jacobi M N;

/* C implementations of common standard library functions. These are *much*
   faster than the originals in stdlib.q and string.q. */

public extern stdlib::append Xs Y, stdlib::cat Xs, stdlib::mklist X N,
  stdlib::nums N M, stdlib::numsby K N M, stdlib::reverse Xs,
  stdlib::tuplecat Xs;

public extern string::chars S, string::join DELIM Xs, string::split DELIM Xs,
  string::strcat Xs;

/* fast (albeit unstable) quicksort implementation using the qsort routine
   from the C library */

public extern sort P Xs;
