
/* tcl/tk interface V0.2 July 2001 AG (some of this code was pilfered from
   Tk_octave) */

/* This file is part of the Q programming system.

   The Q programming system is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   The Q programming system is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#if defined (HAVE_CONFIG_H)
#  include "config.h"
#endif

/* system headers */

#include <stdio.h>
#include <ctype.h>
#include <signal.h>
#include <math.h>

/* check for standard C headers */
#if STDC_HEADERS
# include <stdlib.h>
# include <string.h>
#else
# ifndef HAVE_STRCHR
#  define strchr index
#  define strrchr rindex
# endif
char *strchr (), *strrchr ();
#endif

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif

#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#include <sys/types.h>
#if HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif
#ifndef WEXITSTATUS
# define WEXITSTATUS(stat_val) ((unsigned)(stat_val) >> 8)
#endif
#ifndef WIFEXITED
# define WIFEXITED(stat_val) (((stat_val) & 255) == 0)
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <tcl.h>
#include <tk.h>

#include <libq.h>

/* multithreading support */

#ifdef USE_THREADS
#include <pthread.h>
#endif

#ifdef DMALLOC
#include <dmalloc.h>
#endif

MODULE(tk)

#ifdef USE_THREADS
static pthread_mutex_t tk_mutex;
#endif

Tcl_Interp* interp = NULL;

static void lock_tk(void)
{
#ifdef USE_THREADS
  pthread_mutex_lock(&tk_mutex);
#endif
}

static void unlock_tk(void)
{
#ifdef USE_THREADS
  pthread_mutex_unlock(&tk_mutex);
#endif
}

/* handle X11 protocol errors */

static int XErrorProc(ClientData data, XErrorEvent *errEventPtr)
{
  fprintf(stderr, "X protocol error: ");
  fprintf(stderr, "error=%d request=%d minor=%d\n",
	  errEventPtr->error_code, errEventPtr->request_code,
	  errEventPtr->minor_code);
  return 0;
}

/* handle SIGINT and SIGTERM */

#ifdef _MSC_VER
#define SIGHANDLER_RETURN(status) return
#else
#if RETSIGTYPE == void
#define SIGHANDLER_RETURN(status) return
#else
#define SIGHANDLER_RETURN(status) return status
#endif
#endif

#ifdef MUST_REINSTALL_SIGHANDLERS
#define SIGHANDLER_RESTORE(sig,handler) syssignal(sig,handler)
#else
#define SIGHANDLER_RESTORE(sig,handler) /* nop */
#endif

typedef RETSIGTYPE (*sighandler_t)(int);

static sighandler_t syssignal(sig, handler)
     int sig;
     sighandler_t handler;
{
#ifdef HAVE_POSIX_SIGNALS
  struct sigaction new_action, old_action;
  new_action.sa_handler = handler;
  sigemptyset(&new_action.sa_mask);
  sigemptyset(&old_action.sa_mask);
  new_action.sa_flags = 0;
  sigaction(sig, &new_action, &old_action);
  return old_action.sa_handler;
#else
  return signal(sig, handler);
#endif
}

static int brkflag = 0;
static volatile int brkevent = 0;
static sighandler_t int_handler = NULL, term_handler = NULL,
  hup_handler = NULL;

static RETSIGTYPE break_handler(int sig)
{
  if (sig == SIGINT && int_handler) int_handler(sig);
  if (sig == SIGTERM && term_handler) term_handler(sig);
#ifdef SIGHUP
  if (sig == SIGHUP && hup_handler) hup_handler(sig);
#endif
  SIGHANDLER_RESTORE(sig, break_handler);
  brkevent = 1;
  SIGHANDLER_RETURN(0);
}

/* event source for handling SIGINT/SIGTERM in the Tcl interpreter */

static void break_setup_proc(ClientData data, int flags)
{
}

static int break_event_proc(Tcl_Event *ev, int flags)
{
  brkflag = 1;
  return 1;
}

static void break_check_proc(ClientData data, int flags)
{
  if (brkevent) {
    Tcl_Event *ev = (Tcl_Event*)Tcl_Alloc(sizeof(Tcl_Event));
    if (!ev) return;
    ev->proc = break_event_proc;
    Tcl_QueueEvent(ev, TCL_QUEUE_HEAD);
    brkevent = 0;
  }
}

/* buffer for passing messages from Tcl to Q */

#define CHUNKSZ 200

static char *buf = NULL;
static int bufsz = 0;

static void clear_buf(void)
{
  buf = NULL;
  bufsz = 0;
}

static void init_buf(void)
{
  if (buf) free(buf);
  clear_buf();
}

static int resize_buf(sz)
{
  if (sz >= bufsz) {
    char *newbuf;
    sz = (sz/CHUNKSZ+1)*CHUNKSZ;
    newbuf = buf?realloc(buf, sz):malloc(sz);
    if (!newbuf) return 0;
    bufsz = sz;
    buf = newbuf;
    return 1;
  } else
    return 1;
}

static int add_buf(char *arg)
{
  int argsz = strlen(arg)+1;
  if (bufsz == 0) {
    if (!resize_buf(argsz)) return 0;
    strcpy(buf, arg);
    return 1;
  } else {
    int buflen = strlen(buf);
    if (!resize_buf(argsz+buflen+1)) return 0;
    strcat(buf, " ");
    strcat(buf, arg);
    return 1;
  }
}

/* message queue */

static struct qentry {
  char *s;
  struct qentry *next;
} *head = NULL, *tail = NULL;

static void init_queue(void)
{
  while (head) {
    struct qentry *next = head->next;
    if (head->s) free(head->s);
    free(head);
    head = next;
  }
  head = tail = NULL;
}

static int emptyqueue(void)
{
  return head == NULL;
}

static int enqueue(char *s)
{
  struct qentry *next = malloc(sizeof(struct qentry));
  if (!next) return 0;
  next->s = s;
  next->next = NULL;
  if (head) {
    tail->next = next;
    tail = next;
  } else
    head = tail = next;
  return 1;
}

static char *dequeue(void)
{
  struct qentry *next;
  char *s;
  if (!head) return NULL;
  s = head->s;
  next = head->next;
  free(head);
  if (next)
    head = next;
  else
    head = tail = NULL;
  return s;
}

static int q_send(ClientData clientData,
		  Tcl_Interp *interp,
		  int argc, char **argv)
{
  int i;
  Tcl_ResetResult(interp);
  for (i = 1; i < argc; i++) {
    if (!add_buf(argv[i])) {
      init_buf();
      Tcl_AppendResult(interp, "memory overflow", NULL);
      return TCL_ERROR;
    }
  }
  if (!enqueue(buf)) {
    init_buf();
    Tcl_AppendResult(interp, "memory overflow", NULL);
    return TCL_ERROR;
  }
  clear_buf();
  return TCL_OK;
}

/* operations dealing with the Tk interpreter */

static void tk_stop(void)
{
  if (interp) {
    Tcl_DeleteInterp(interp);
    interp = NULL;
    init_buf();
    init_queue();
  }
}

static void tk_do_events(int sync)
{
  if (!interp) return;
  if (Tk_GetNumMainWindows() > 0 && sync) Tcl_DoOneEvent(0);
  while(Tk_GetNumMainWindows() > 0 && Tcl_DoOneEvent(TCL_DONT_WAIT));
  if (Tk_GetNumMainWindows() == 0) tk_stop();
}

static int tk_running(void)
{
  tk_do_events(0);
  return interp != NULL;
}

static int tk_chk(void)
{
  tk_do_events(0);
  return !emptyqueue();
}

static char *tk_recv(void)
{
  brkflag = 0;
  tk_do_events(0);
  while (interp && Tk_GetNumMainWindows() > 0 && !brkflag && emptyqueue()) {
    release_lock();
    tk_do_events(1);
    acquire_lock();
  }
  brkflag = 0;
  if (emptyqueue())
    return NULL;
  else
    return dequeue();
}

static char *result = NULL;

static void set_result(char *s)
{
  result = malloc(strlen(s)+1);
  if (result) strcpy(result, s);
}

static int tk_eval(char *s)
{
  int status;
  char *cmd;
  result = NULL;
  if (!interp) return 0;
  cmd = malloc(strlen(s)+1);
  if (!cmd) return 0;
  strcpy(cmd, s);
  status = Tcl_Eval(interp, cmd);
  if (interp && interp->result && *interp->result)
    set_result(interp->result);
  else if (status == TCL_BREAK)
    set_result("invoked \"break\" outside of a loop");
  else if (status == TCL_CONTINUE)
    set_result("invoked \"continue\" outside of a loop");
  else
    set_result("");
  if (status == TCL_BREAK || status == TCL_CONTINUE)
    status = TCL_ERROR;
  tk_do_events(0);
  free(cmd);
  return status != TCL_ERROR;
}

static int tk_start(void)
{
  Tk_Window mainw;
  result = NULL;
  if (interp) return 1;
  /* start up a new interpreter */
  if (!(interp = Tcl_CreateInterp())) return 0;
  if (Tcl_Init(interp) != TCL_OK) {
    if (interp->result && *interp->result)
      set_result(interp->result);
    else
      set_result("error initializing Tcl");
    tk_stop();
    return 0;
  }
  /* intialize message queue for communication between Q and Tcl */
  init_buf();
  init_queue();
  /* create command to send messages from Tcl to Q */
  Tcl_CreateCommand(interp, "q", q_send, NULL, NULL);
  /* oddly, there are no `env' variables passed, and this one is needed */
  Tcl_SetVar2(interp, "env", "DISPLAY", getenv("DISPLAY"), TCL_GLOBAL_ONLY);
  if (Tk_Init(interp) != TCL_OK) {
    if (interp->result && *interp->result)
      set_result(interp->result);
    else
      set_result("error initializing Tk");
    tk_stop();
    return 0;
  }
  /* set up an X error handler */
  mainw = Tk_MainWindow(interp);
  Tk_CreateErrorHandler(Tk_Display(mainw), -1, -1, -1,
			XErrorProc, (ClientData)mainw);
  return 1;
}

FUNCTION(tk,tk,argc,argv)
{
  char *s;
  if (argc == 1 && isstr(argv[0], &s)) {
    if (tk_start()) {
      int res = tk_eval(s);
      if (!result)
	return __ERROR;
      else if (res)
	if (*result)
	  return mkstr(result);
	else {
	  free(result);
	  return mkvoid;
	}
      else
	return mkapp(mksym(sym(tk_error)), mkstr(result));
    } else if (result)
      return mkapp(mksym(sym(tk_error)), mkstr(result));
    else
      return __FAIL;
  } else
    return __FAIL;
}

FUNCTION(tk,tk_quit,argc,argv)
{
  if (argc == 0) {
    tk_stop();
    return mkvoid;
  } else
    return __FAIL;
}

FUNCTION(tk,tk_ready,argc,argv)
{
  if (argc == 0) {
    if (tk_running())
      return mktrue;
    else
      return mkfalse;
  } else
    return __FAIL;
}

FUNCTION(tk,tk_check,argc,argv)
{
  if (argc == 0) {
    if (tk_chk())
      return mktrue;
    else
      return mkfalse;
  } else
    return __FAIL;
}

FUNCTION(tk,tk_reads,argc,argv)
{
  if (argc == 0) {
    char *s;
    lock_tk();
    s = tk_recv();
    unlock_tk();
    if (s)
      return mkstr(s);
    else
      return __FAIL;
  } else
    return __FAIL;
}

INIT(tk)
{
#ifdef USE_THREADS
  pthread_mutex_init(&tk_mutex, NULL);
#endif
  /* register signal handlers and event sources */
  int_handler = syssignal(SIGINT, break_handler);
  term_handler = syssignal(SIGTERM, break_handler);
#ifdef SIGHUP
  hup_handler = syssignal(SIGHUP, break_handler);
#endif
  Tcl_CreateEventSource(break_setup_proc, break_check_proc, NULL);
}

FINI(tk)
{
  /* tk_stop(); */
  Tcl_Finalize();
}
