
/* except.q: exception handling for the builtin and standard library functions
   03-04-02 AG (orginal version distributed as the example "except" script,
   07-26-01 AG) */

/* This file is part of the Q programming system.

   The Q programming system is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   The Q programming system is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/* This module provides default rules for the builtin and standard library
   operations, which throw an exception whenever such an operation fails. The
   exception handling implemented here is fairly simplistic: just throw back
   the offending expression protected with a special constructor. There are
   different constructors, belonging to different Exception subtypes, for the
   builtins and each part of the standard library, such as builtin_err of type
   BuiltinException for the builtins, stdlib_err of type StdlibException for
   the operations in stdlib.q, etc.

   There also is a simple handler function which can be supplied as the first
   argument to `catch'. When invoked with an exception, i.e., a value of type
   Exception, it prints an appropriate error message and halts. Otherwise it
   just returns the argument value unchanged.

   Examples: Default error message printed by the interpreter in response to
   an unhandled exception:

   ==> 17+4*3/0
   ! Exception
   builtin_err (12/0)
   >>> 17+4*3/0
               ^

   Use our own exception handler:

   ==> catch exception (17+4*3/0)
   ! Exception: builtin_err (12/0)

   You can also handle exceptions in the Q machine. E.g., let's force a stack
   overflow:

   ==> catch exception (nums 1 100000000)
   ! Exception: syserr 5

   Note that 5 is the error code for an expression stack overflow; see the
   manual for the list of all built-in exception codes.  */

public exception X;

exception X:Exception	= fwrites ERROR ("! Exception: "++str X++"\n") ||
			  halt;
exception X		= X otherwise;

/* The following rules are at the lowest possible priority level so as not to
   interfere with the standard library and user equations. */

@-0x80000000

/* Note that some operations are missing in this list, because they never fail
   unless a hard error condition occurs. */

/* builtins ******************************************************************/

public type BuiltinException : Exception = special const builtin_err X;

/* operators */

X^Y			= throw (builtin_err (X^Y));
X!Y			= throw (builtin_err (X!Y));

-X			= throw (builtin_err (-X));
#X			= throw (builtin_err (#X));
not X			= throw (builtin_err (not X));

X*Y			= throw (builtin_err (X*Y));
X/Y			= throw (builtin_err (X/Y));
X div Y			= throw (builtin_err (X div Y));
X mod Y			= throw (builtin_err (X mod Y));
X and Y			= throw (builtin_err (X and Y));
X and then Y		= throw (builtin_err (X and then Y));

X++Y			= throw (builtin_err (X++Y));
X+Y			= throw (builtin_err (X+Y));
X-Y			= throw (builtin_err (X-Y));
X or Y			= throw (builtin_err (X or Y));
X or else Y		= throw (builtin_err (X or else Y));

X<Y			= throw (builtin_err (X<Y));
X>Y			= throw (builtin_err (X>Y));
(X=Y)			= throw (builtin_err (X=Y));
X<=Y			= throw (builtin_err (X<=Y));
X>=Y			= throw (builtin_err (X>=Y));
X<>Y			= throw (builtin_err (X<>Y));

/* functions */

succ X			= throw (builtin_err (succ X));
pred X			= throw (builtin_err (pred X));

shl N COUNT		= throw (builtin_err (shl N COUNT));
shr N COUNT		= throw (builtin_err (shr N COUNT));

exp X			= throw (builtin_err (exp X));
ln X			= throw (builtin_err (ln X));
sqrt X			= throw (builtin_err (sqrt X));
sin X			= throw (builtin_err (sin X));
cos X			= throw (builtin_err (cos X));
atan X			= throw (builtin_err (atan X));
atan2 X Y		= throw (builtin_err (atan2 X Y));
seed N			= throw (builtin_err (seed N));


sub Xs I J		= throw (builtin_err (sub Xs I J));
substr S K L		= throw (builtin_err (substr S K L));
pos S1 S		= throw (builtin_err (pos S1 S));

trunc X			= throw (builtin_err (trunc X));
round X			= throw (builtin_err (round X));
float X			= throw (builtin_err (float X));
int X			= throw (builtin_err (int X));
frac X			= throw (builtin_err (frac X));
ord X			= throw (builtin_err (ord X));
chr N			= throw (builtin_err (chr N));
list Xs			= throw (builtin_err (list Xs));
tuple Xs		= throw (builtin_err (tuple Xs));
str X			= throw (builtin_err (str X));
strq X			= throw (builtin_err (strq X));
val S			= throw (builtin_err (val S));
valq S			= throw (builtin_err (valq S));

fopen NAME MODE		= throw (builtin_err (fopen NAME MODE));
popen CMD MODE		= throw (builtin_err (popen CMD MODE));
fclose F		= throw (builtin_err (fclose F));
read			= throw (builtin_err read);
readq			= throw (builtin_err readq);
readc			= throw (builtin_err readc);
reads			= throw (builtin_err reads);
fread F			= throw (builtin_err (fread F));
freadq F		= throw (builtin_err (freadq F));
freadc F		= throw (builtin_err (freadc F));
freads F		= throw (builtin_err (freads F));
write X			= throw (builtin_err (write X));
writeq X		= throw (builtin_err (writeq X));
writec X		= throw (builtin_err (writec X));
writes S		= throw (builtin_err (writes S));
fwrite F X		= throw (builtin_err (fwrite F X));
fwriteq F X		= throw (builtin_err (fwriteq F X));
fwritec F X		= throw (builtin_err (fwritec F X));
fwrites F S		= throw (builtin_err (fwrites F S));
eof			= throw (builtin_err eof);
feof F			= throw (builtin_err (feof F));
flush			= throw (builtin_err flush);
fflush F		= throw (builtin_err (fflush F));

time			= throw (builtin_err time);
sleep X			= throw (builtin_err (sleep X));

/* complex *******************************************************************/

import complex;

public type ComplexException : Exception = special const complex_err X;

arg Z			= throw (complex_err (arg Z));
re Z			= throw (complex_err (re Z));
im Z			= throw (complex_err (im Z));
conj Z			= throw (complex_err (conj Z));

/* cond **********************************************************************/

import cond;

public type CondException : Exception = special const cond_err X;

ifelse P X Y		= throw (cond_err (ifelse P X Y));
switch CASES		= throw (cond_err (switch CASES));
match X CASES		= throw (cond_err (match X CASES));

/* error *********************************************************************/

import error;

public type ErrorException : Exception = special const error_err X;

error S			= throw (error_err (error S));

/* lambda ********************************************************************/

import lambda;

public type LambdaException : Exception = special const lambda_err X;

lambda X Y		= throw (lambda_err (lambda X Y));

_H X Y			= throw (lambda_err (_H X Y));
_HL X Y			= throw (lambda_err (_HL X Y));
_HT X Y			= throw (lambda_err (_HT X Y));
_HK X Y Z		= throw (lambda_err (_HK X Y Z));
__H X Y			= throw (lambda_err (__H X Y));
__HL X Y		= throw (lambda_err (__HL X Y));
__HT X Y		= throw (lambda_err (__HT X Y));
__HK X Y Z		= throw (lambda_err (__HK X Y Z));

/* list **********************************************************************/

/* none */

/* math **********************************************************************/

import math;

public type MathException : Exception = special const math_err X;

asin X			= throw (math_err (asin X));
acos X			= throw (math_err (acos X));
tan X			= throw (math_err (tan X));
lg X			= throw (math_err (lg X));
log X			= throw (math_err (log X));
sinh X			= throw (math_err (sinh X));
cosh X			= throw (math_err (cosh X));
tanh X			= throw (math_err (tanh X));
asinh X			= throw (math_err (asinh X));
acosh X			= throw (math_err (acosh X));
atanh X			= throw (math_err (atanh X));

/* sort **********************************************************************/

import sort;

public type SortException : Exception = special const sort_err X;

msort P Xs		= throw (sort_err (msort P Xs));
qsort P Xs		= throw (sort_err (qsort P Xs));

/* stdtypes ******************************************************************/

import stdtypes;

public type StdtypesException : Exception = special const stdtypes_err X;

array Xs		= throw (stdtypes_err (array Xs));
array2 Xs		= throw (stdtypes_err (array2 Xs));
bag Xs			= throw (stdtypes_err (bag Xs));
delete A X		= throw (stdtypes_err (delete A X));
dict XYs		= throw (stdtypes_err (dict XYs));
first A			= throw (stdtypes_err (first A));
heap Xs			= throw (stdtypes_err (heap Xs));
insert A X		= throw (stdtypes_err (insert A X));
keys D			= throw (stdtypes_err (keys D));
last A:Array		= throw (stdtypes_err (last A));
last M:Bag		= throw (stdtypes_err (last M));
last D:Dict		= throw (stdtypes_err (last D));
last M:Set		= throw (stdtypes_err (last M));
member A X		= throw (stdtypes_err (member A X));
members A		= throw (stdtypes_err (members A));
members2 A		= throw (stdtypes_err (members2 A));
mkarray X N		= throw (stdtypes_err (mkarray X N));
mkarray2 X NM		= throw (stdtypes_err (mkarray2 X NM));
mkdict Y Xs		= throw (stdtypes_err (mkdict Y Xs));
rmfirst A		= throw (stdtypes_err (rmfirst A));
rmlast A		= throw (stdtypes_err (rmlast A));
set Xs			= throw (stdtypes_err (set Xs));
update A K X		= throw (stdtypes_err (update A K X));
update2 A IJ X		= throw (stdtypes_err (update2 A IJ X));
vals D			= throw (stdtypes_err (vals D));

/* stream ********************************************************************/

import stream;

public type StreamException : Exception = special const stream_err X;

stream Xs		= throw (stream_err (stream Xs));
foldl1 F Xs:Stream	= throw (stream_err (foldl1 F Xs));
foldr1 F Xs:Stream	= throw (stream_err (foldr1 F Xs));
hd Xs:Stream		= throw (stream_err (hd Xs));
init Xs:Stream		= throw (stream_err (init Xs));
last Xs:Stream		= throw (stream_err (last Xs));
pop Xs:Stream		= throw (stream_err (pop Xs));
tl Xs:Stream		= throw (stream_err (tl Xs));
top Xs:Stream		= throw (stream_err (top Xs));
numstream N		= throw (stream_err (numstream N));
numstreamby K N		= throw (stream_err (numstreamby K N));

/* string ********************************************************************/

import string;

public type StringException : Exception = special const string_err X;

chars S			= throw (string_err (chars S));
join DELIM Xs		= throw (string_err (join DELIM Xs));
split DELIM S		= throw (string_err (split DELIM S));
strcat Xs		= throw (string_err (strcat Xs));

/* stdlib ********************************************************************/

/* this module comes last because some of the default rules here are
   overridden above (overloaded operations!) */

import stdlib;

public type StdlibException : Exception = special const stdlib_err X;

abs X			= throw (stdlib_err (abs X));
all P Xs		= throw (stdlib_err (all P Xs));
any P Xs		= throw (stdlib_err (any P Xs));
append Xs Y		= throw (stdlib_err (append Xs Y));
cat Xs			= throw (stdlib_err (cat Xs));
cons X Xs		= throw (stdlib_err (cons X Xs));
do F Xs			= throw (stdlib_err (do F Xs));
drop N Xs		= throw (stdlib_err (drop N Xs));
dropwhile P Xs		= throw (stdlib_err (dropwhile P Xs));
filter P Xs		= throw (stdlib_err (filter P Xs));
foldl F A Xs		= throw (stdlib_err (foldl F A Xs));
foldl1 F Xs		= throw (stdlib_err (foldl1 F Xs));
foldr F A Xs		= throw (stdlib_err (foldr F A Xs));
foldr1 F Xs		= throw (stdlib_err (foldr1 F Xs));
fst Xs			= throw (stdlib_err (fst Xs));
hd Xs			= throw (stdlib_err (hd Xs));
hds Xss			= throw (stdlib_err (hds Xss));
init Xs			= throw (stdlib_err (init Xs));
iter N F A		= throw (stdlib_err (iter N F A));
last Xs			= throw (stdlib_err (last Xs));
map F Xs		= throw (stdlib_err (map F Xs));
mklist X N		= throw (stdlib_err (mklist X N));
null Xs			= throw (stdlib_err (null Xs));
nums N M		= throw (stdlib_err (nums N M));
numsby K N M		= throw (stdlib_err (numsby K N M));
pop Xs			= throw (stdlib_err (pop Xs));
prd Xs			= throw (stdlib_err (prd Xs));
push Xs X		= throw (stdlib_err (push Xs X));
reverse Xs		= throw (stdlib_err (reverse Xs));
scan F A Xs		= throw (stdlib_err (scan F A Xs));
scan1 F Xs		= throw (stdlib_err (scan1 F Xs));
sgn X			= throw (stdlib_err (sgn X));
snd Xs			= throw (stdlib_err (snd Xs));
sum Xs			= throw (stdlib_err (sum Xs));
take N Xs		= throw (stdlib_err (take N Xs));
takewhile P Xs		= throw (stdlib_err (takewhile P Xs));
tl Xs			= throw (stdlib_err (tl Xs));
tls Xss			= throw (stdlib_err (tls Xss));
top Xs			= throw (stdlib_err (top Xs));
transpose Xss		= throw (stdlib_err (transpose Xss));
trd Xs			= throw (stdlib_err (trd Xs));
tuplecat Xs		= throw (stdlib_err (tuplecat Xs));
unzip Xs:List		= throw (stdlib_err (unzip Xs:List));
unzip3 Xs:List		= throw (stdlib_err (unzip3 Xs:List));
zip Xs Ys		= throw (stdlib_err (zip Xs Ys));
zip3 Xs Ys Zs		= throw (stdlib_err (zip3 Xs Ys Zs));
zipwith F Xs Ys		= throw (stdlib_err (zipwith F Xs Ys));
zipwith3 F Xs Ys Zs	= throw (stdlib_err (zipwith3 F Xs Ys Zs));

/* clib **********************************************************************/

include clib_except;
